from sage.all import *
import sys

########################
### global variables ###
########################
zero = 0
field = []
verbose = False
variables = 'x'
size = 0
rack = []
q = []
monomial = []
derivative = []
info = {}

############
### show ###
############
def show():
    print 'Field=%s' %field
    print 'Size=%d' %size;
    print 'Rack';
    for i in range(0,size): 
        for j in range(0,size):
            print rack[i][j],
        print;

    print '2-Cocycle';
    for i in range(0,size): 
        for j in range(0,size):
            print q[i][j],
        print;
    if derivative==[]:
        print 'The file contains no sequence of derivatives'
    else:
        derivative.reverse()
        print 'Derivative=%s'% derivative

    return

#################
### read_file ###
#################
def read_file(filename):
    global rack, q, monomial, size, derivative, zero, field

    file = open(filename, 'r');
    buffer = "";

    for line in file.readlines():
            for i in range(0, len(line)):
                if (line[i] == '#'):
                    buffer = buffer + '\n'
                    break;
                if (line[i] != ' '):
                    buffer = buffer + line[i];

    newbuffer = buffer.split('\n');
    for line in newbuffer:
        if (line.find('monomial=') != -1):
            tmp = line.split('monomial=');
            m = str(tmp[1]);
            monomial = {tuple(map(int, m[1:-1].split(','))):1}

        if (line.find('characteristic=') != -1):
            tmp = line.split('characteristic=');
            char = int(tmp[1])

        if (line.find('derivative=') != -1):
            tmp = line.split('derivative=');
            d = str(tmp[1])
            derivative = map(int, d.split(','))
    # rack 
    if 'rack' in newbuffer:
        position = newbuffer.index('rack');

        # size
        tmp = newbuffer[position+1].split(',')
        size = len(tmp)

        rack = [];
        rack = [[0 for i in range(size)] for j in range(size)]
        q = [];
        q = [[0 for i in range(size)] for j in range(size)]
        for i in range(0, size):
            tmp = newbuffer[position+i+1].split(',');
            for j in range(0,size):
                rack[i][j] = int(tmp[j]);

    # 2-cocycle					
    if 'q' in newbuffer:
        position = newbuffer.index('q');
        for i in range(0, size):
            tmp = newbuffer[position+i+1].split(',');
            for j in range(0,size):
                try:
                    q[i][j] = int(tmp[j])
                except:
                    q[i][j] = gap.new(tmp[j])

        field = gap.DefaultFieldOfMatrix(q)
        if gap.IsFinite(field):
            print 'Working over the field', field
        else:
            print 'Working over the complex numbers'


######################
### count_elements ###
######################
def count_elements(p):
    """ ... """
    c = []
    for i in range(len(rack)):
        c = c + [[0,i]]
    for e in p:
        for i in range(len(e)):
            c[e[i]][0] += 1
    c.sort()
    while c[0][0] == 0:
        del c[0]
    return c

###############
### is_zero ###
###############
def is_zero(p, bound=0):
    return not is_not_zero(p,bound)

###################
### is_not_zero ###
###################
def is_not_zero(p, bound=0):
    """ 
    Check if a polynomial is zero or not 
    
    """
    if not p:
        return False

    if p.keys() == [()]:
        return True

    c = count_elements(p)

    for i in c[:bound*(bound!=0)+len(c)*(bound==0)]:
        if verbose:
            print "Derivative %d, degree %d" %(i[1], len(p.keys()[0]))

        info[len(p.keys()[0])] = i[1]
        tmp = derive(i[1], p)
        if is_not_zero(tmp, bound):
            return True
        
    return False

##############
### derive ###
##############
def derive(i, p):
    """ 
    Derive a polynomial of the form [a,[i,j,...k]] + [b,[r,s,...,t]] + ... 
    Also it is possible to use for successive derivations.

    """
    if type(i) == tuple:
        tmp = p
        for k in range(0, len(i)):
            tmp = derive(i[k], tmp)
    else:
        tmp = dict()
        for m in p.keys():
            v = derive_monomial(i, m)
            for k in v.keys():
                if k in tmp:
                    tmp[k] += p[m]*v[k]
                else:
                    tmp[k] = p[m]*v[k]

                ### works for non-zero characteristic 
                zero = gap.Zero(field)
                if tmp[k] == zero:
                    del tmp[k]

    return tmp

#######################
### derive_monomial ###
#######################
def derive_monomial(i, m):
    """ derive a monomial of the form [i,j,...,k] """

    p = dict()
    s = 0
    n = len(m)
    
    if n == 1: 
        if i==m[0]:
            p[()] = 1
        return p

    for j in range(0,n):
        tmp = ();
        cof = 1;
        if i == m[j]:
            s = s+1
            if j-1 >= 0:
                for k in range(0,j):
                    tmp = tmp + (m[k],)

            for k in range(j+1,n):
                tmp = tmp + (rack[i][m[k]],)
                cof = cof * q[i][m[k]]

            if tmp in p:
                p[tmp] = p[tmp] + cof
            else:
                p[tmp] = cof

    return p

############
### help ###
############
def help():
    print 'Usage: sage derSAGE.py [OPTION] [FILE]'
    print '  -v, --verbose         explain what is being done'
    print '  -h, --help            display this help and exit'

############
### main ###
############
def main():
    global verbose
    if len(sys.argv) == 1:
        help()
        sys.exit()

    if ('-v' in sys.argv) or ('--verbose' in sys.argv):
        verbose = True
        try:
            sys.argv.remove('-v')
        except:
            sys.argv.remove('--verbose')

    if (len(sys.argv) == 1) or ('-h' in sys.argv) or ('--help' in sys.argv):
        help()
        sys.exit(0)

    read_file(sys.argv[1])
    show()

    if monomial == "":
        sys.exit(1)
    else:
        print 'Monomial=%s' % monomial
        if derivative == []:
            print 'Checking if the monomial is zero'
            if is_zero(monomial) == True:
                print '%s is zero!' % monomial
            else:
                print '%s is not zero' % monomial
                sq = [info[x+1] for x in range(len(info))]
                print 'Sequence of derivatives: %s' % sq
 
        else:
            m = monomial
            print 'Derivative %s' % derivative
            for d in derivative:
                m = derive(d, m)
                print 'D(%d) is %s'%(d,m)

if __name__ == "__main__":
    main()

# vim:ts=4:sw=4:expandtab:
